/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */


//           Routines for printing structures and debugging.

#include "Tree.h"

void tree::printlists(int level,int start,int u,int st,int ed) const{
  //Mostly for debugging.
  cout<<"tree number "<<start<<" at level"<<level<<":\n\n";
  cout<<*this<<"\n\n";
  cout<<"down lists:\n\n";
  for(int i=0;i<NumCodons;i++){
    cout<<*(down+i)<<"  ";
  };
  cout<<"\n\n";
  if(ed==0){
    ed=this->edges();
  };
  if(u){
    cout<<"up lists:\n\n";
    for(int i=0;i<NumCodons;i++){
      cout<<*(up+i)<<"  ";
    };
    cout<<"\n\n";
  };
  cout<<"extra lists:\n\n";
  for(int j=0;j<params::numpars;j++){
    for(int i=0;i<NumCodons;i++){
      cout<<*(extra+j*NumCodons+i)<<"  ";
    };
    cout<<"\n\n";
  };
  cout<<"Numbers "<<st<<" to "<<st+this->edges()-1<<":\n\n";
  for(int j=st;j<st+this->edges();j++){
    for(int i=0;i<NumCodons;i++){
      cout<<*(extra+(params::numpars+j)*NumCodons+i)<<"  ";
    };
    cout<<"\n\n";
  };
  cout<<"\n\n";
  int e=0;
  for(int i=0;i<n;i++){
    (subtree+i)->printlists(level+1,i,1,st+e+1,ed);
    e+=(subtree+i)->edges()+1;
  };
};


void tree::debugneg(long double val,ostream& out,int siteno,int l){
  //Called when the likelihood function for the current site would
  //give a non-positive number. It attempts to determine the cause
  //and tries to determine where the problem occurs.
  if(l==0){//Root node
    out<<"Likelihood function for ";
    if(siteno==-1){
      out<<"unknown site";
    }else{
      out<<"site number "<<siteno+1;
    };
    out<<" gives invalid value: "<<val<<".\n";
  };
  int problem=0;//all values zero.
  for(int i=0;i<NumCodons;i++){
    if(*(down+i)<0){//Oh dear!
      problem=1;
      break;
    };
    if(*(down+i)>0){
      problem=2;
    };
  };
  switch(problem){
  case 0://All values zero. Almost certainly caused by underflow.
    out<<"Probable Underflow.\n";
    break;
  case 1://Some negative values. Possibly caused by rounding errors?
    out<<"Negative values encountered:\nCodon\tvalue\n";
    for(int i=0;i<NumCodons;i++){
      if(*(down+i)<0){
	out<<i<<"\t"<<*(down+i)<<"\n";
      };
    };
    out<<"\n";
    break;
  case 2://Good values. Problem must be caused by underflow when
	 //multiplying by pi.
    out<<"Values in likelihood list seem valid. Problem most likely caused by errors in the Q-Matrix.\n";
    out<<"Values:\n";
    for(int i=0;i<NumCodons;i++){
      out<<*(down+i)<<"\t";
    };
    out<<"\n\n";
    return;
  };
  problem=0;
  for(int i=0;i<n;i++){//test subtrees for root of problem.
    int prob=0;
    for(int j=0;j<NumCodons;j++){
      if(*((subtree+i)->down2+j)<0){//Oh dear!
	prob=1;
	break;
      };
      if(*((subtree+i)->down2+j)>0){
	prob=2;
      };
    };
    if(prob!=2){//problem in this subtree.
      out<<"Problem comes from subtree number "<<i<<" at level "<<l+1<<".\n";
      for(int j=0;j<NumCodons;j++){
	cout<<*((subtree+i)->down2+j)<<"\t";
      };
      cout<<"\n";
      (subtree+i)->debugneg(val,out,siteno,l+1);
      problem=prob;
    };
  };
  if(problem==0&&n>0){//all subtrees OK
    out<<"All subtrees seem OK. Probable underflow in combining them.\n";
  };
  if(n==0){//Problem assigning codon likelihoods
    out<<"Data corrupted. Likelihood list at leaf:\n";
    for(int i=0;i<NumCodons;i++){
      out<<*(down+i)<<"\t";
    };
    out<<"\n\n";
  };
};


ostream &operator <<(ostream &o,const tree &T){
  if(T.n==0){
    o<<T.label;
    return o;
  };
  o<<'(';
  for(int i=0;i<T.n;i++){
    o<<*(T.subtree+i);
    o<<':'<<*(T.length+i);
    if(i<T.n-1){
      o<<',';
    };
  };
  o<<')';
  return o;
};

void tree::printdata(ostream& out, sequence* d) const{
  if(n==0){
    out<<*d<<"\n";
  }else{
    int pos=0;
    for(int i=0;i<n;i++){
      (subtree+i)->printdata(out,d+pos);
      pos+=(subtree+i)->leaves();
    };
  };
};

void tree::printspecies(ostream& out,int* no) const{
  if(n==0){
    if(*no){
      out<<label<<"\n";
    };
  }else{
    for(int i=0;i<n;i++){
      (subtree+i)->printspecies(out,no);
      no+=(subtree+i)->leaves();
    };
  };
};


//Output Routines

void tree::print(ostream& out) const{
  if(n==0){
    //    out<<"\""<<label<<"\"";
    out<<label;
    return;
  };
  out<<'(';
  for(int i=0;i<n;i++){
    (subtree+i)->print(out);
    out<<':'<<*(length+i);
    if(i<n-1){
      out<<',';
    };
  };
  out<<')';
};

void tree::print(long double *bl,ostream& out) const{
  if(n==0){
    out<<label;
    return;
  };
  out<<'(';
  for(int i=0;i<n;i++){
    (subtree+i)->print(bl+1,out);
    out<<':'<<*bl;
    if(i<n-1){
      out<<',';
      bl+=(subtree+i)->numed+1;
    };
  };
  out<<')';
};
