/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#include <iostream>
#include <fstream>
#include <stdlib.h>
#include <string>
#include <math.h>
#include <time.h>
using namespace std;

ofstream commentry ("/dev/null");
ostream& comm=cout;//commentry;

#include "ErrorHandler.h"
#include "SignalHandler.h"
#include "CommandLine.h"
extern varlist vars;
#include "Matrix.h"
#include "Sequence.h"
#include "Parameters.h"
#include "Likelihood.h"
#include "Optimisation.h"
#include "Tree.h"
#include "TreeLikelihood.h"
#include "MixTreeLikelihood.h"
#include "Miscellaneous.h"
#include "Macros.h"
#include "Input.h"
#include "fileExtensions.h"

void testundopars(int *sel,ifstream& Qmat){
  params p;
  params::orthmats();
  variable *rec=vars.seektag("-initpars");
  if(rec!=NULL){
    char *x=new char[50+strlen(*(char **)rec->value)];
    sprintf(x,"Reading parameters from file \"%s\".",*(char**)rec->value);
    info(x,msgcode(1,0));
    ifstream in;
    setifstream(*(char**)rec->value,&in);
    if(!in.good()){
      sprintf(x,"Could not open initial parameter file \"%s\"",*(char**)rec->value);
      recoverableError(x);//should give file menu.
    };
    p.readpars(in);
    in.close();
    delete[] x;
  }else{
    Realmatrix Q(Qmat);
    p.getpars(Q);
  };
  p.normalise();
  long double *tc=p.truepars();
  for(int i=0;i<params::numpars;i++){
    cout<<*(tc+i)<<" ";
  };
  cout<<"\n\n";
  int np=params::numpars;
  Realmatrix *oldx=new Realmatrix[np];
  for(int i=0;i<np;i++){
    *(oldx+i)=*(params::XX+i);
  };
  p.selectpars(sel);
  p.undoselect(sel,np);
  tc=p.truepars();
  for(int i=0;i<params::numpars;i++){
    cout<<*(tc+i)<<" ";
  };
  cout<<"\n\n";
  for(int i=0;i<np;i++){
    long double err=0;
    for(int j=0;j<NumCodonsSq;j++){
      long double df=*((oldx+i)->entries+j)-*((params::XX+i)->entries+j);
      if(df<0){
	df=-df;
      };
      if(df>err){
	err=df;
      };
    };
    cout<<err<<"\n";
  };
};

void simulate(const char* basefilename,long unsigned int numfiles, long unsigned int seqlength,const tree &t,params &p);

inline int getskip(char *ret){
  //changes return point into a suitable number, allowing certain
  //calculations to be skipped when resuming from a partial calculation.
  if(!strcmp(ret,"newx")){
    return 1;
  }else if(!strcmp(ret,"hessian")){
    return 2;
  }else if(!strcmp(ret,"endhessian")){
    return 3;
  }else if(!strcmp(ret,"newx")){
    return 1;
  }else if(!strcmp(ret,"newx")){
    return 1;
  }else if(!strcmp(ret,"newx")){
    return 1;
  }else if(!strcmp(ret,"newx")){
    return 1;
  }else{
    warning("Unable to identify saved state - recovery file may be incorrect or corrupted.");
    endrecover();
    return 0;
  };
};

void setmodels(){
  const char *modelfile=MAKESTRING(DEFAULTMODELFILE);
  variable *rec=vars.seektag("-modelfile");
  if(rec!=NULL){//Use non-standard model file.
    modelfile=*(char **)rec->value;  
  };
  rec=vars.seektag("-model");
  if(rec!=NULL){
    const char *mdl=*(char**)rec->value;
    int mnamelen=0;
    const char *mdlps=mdl;
    for(;*mdlps&&*mdlps!='(';mdlps++){//Need to add support for quotes
      mnamelen++;
      if(*mdlps=='\\'){
	mdlps++;
	if(!*mdlps){
	  break;
	};
      };
    };
    char *mname=new char[mnamelen+1];
    const char *mdlpos2=mdl;
    for(int i=0;i<mnamelen;i++){
      if(*mdlpos2=='\\'){
	mdlpos2++;
	if(!*mdlpos2){
	  break;
	};
      };
      *(mname+i)=*(mdlpos2++);
    };
    *(mname+mnamelen)='\0';
    for(;*mdlps&&*mdlps!='(';mdlps++);
    mdlpos2=mdlps;
    int npars=0;
    for(;*mdlps;mdlps++){
      if(*mdlps==','){
	npars++;
      };
      if(*mdlps==')'){
	npars++;
	break;
      };
      if(*mdlps=='\\'){
	mdlps++;
	if(!*mdlps){
	  break;
	};
      };
    };
    mdlps=++mdlpos2;
    char **vvals=new char *[npars];
    for(int i=0;i<npars;i++){
      int len=0;
      for(;*mdlpos2&&*mdlpos2!=','&&*mdlpos2!=')';mdlpos2++){
	if(*mdlpos2=='\\'){
	  mdlpos2++;
	  if(!*mdlpos2){
	    break;
	  };
	};
	len++;
      };
      *(vvals+i)=new char[len+1];
      char *vvcur=*(vvals+i);
      for(;*mdlps&&*mdlps!=','&&*mdlps!=')';mdlps++){
	if(*mdlps=='\\'){
	  mdlps++;
	  if(!*mdlps){
	    break;
	  };
	};
	*(vvcur++)=*mdlps;
      };
      *vvcur='\0';
      mdlps++;
      mdlpos2++;
    };
    ifstream modelstream;
    setifstream(modelfile,&modelstream);
    readmodel(modelstream,mname,npars,vvals);
    modelstream.close();
    for(int i=0;i<npars;i++){
      delete[] *(vvals+i);
    };
    delete[] vvals;
    delete[] mname;
  };
};

void checkformask(){
  ifstream mask;
  variable *rec=vars.seektag("-maskfile");
  if(rec!=NULL){
    setifstream(*(char**)rec->value,&mask);
    if(!mask.is_open()){
      char *msg=new char[50+strlen(*(char**)rec->value)];
      strcpy(msg,"Unable to open mask file \"");
      strcat(msg,*(char**)rec->value);
      strcat(msg,"\"");
      warning(msg);
      delete[] msg;
    };
  }else{
    setifstream(MAKESTRING(DEFAULTMASKFILE),&mask);    
  };
  rec=vars.seektag("-mask");
  if(rec!=NULL){
    int msk=atoi(*(char**)rec->value);
    if(!mask.is_open()){
      recoverableError("Unable to open mask file.");
    };
    params::getmask(msk,mask);
  };
  mask.close();
};

treelk* maketreelikelihood(char *a,ifstream& data,ifstream& Qmat,int skip){
  treelk* ff;
  int num=1;
  variable *rec=vars.seektag("-mixture");
  if(rec!=NULL){
    num=atoi(*(char**)rec->value);
  };
  if(num>1){
    ff=new Mixtreelikelihood(num,a,data);
    long double *pr=new long double[num-1];
    rec=vars.seektag("-fixedprobs");
    if(rec==NULL){
      rec=vars.seektag("-initprobs");
      if(rec==NULL){
	for(int i=0;i<num-1;i++){
	  *(pr+i)=1; 
	};
	((Mixtreelikelihood *)ff)->setprob(pr);
      }else{
	const char *pbs=*(char **)rec->value;
	long double lstp=1;
	for(int i=0;i<num-1;i++){
	  *(pr+i)=atof(pbs);
	  lstp-=*(pr+i);
	  for(;*pbs&&*pbs!=',';pbs++);
	  pbs++;
	};
	for(int i=0;i<num-1;i++){
	  *(pr+i)/=lstp;
	};
	((Mixtreelikelihood *)ff)->setprob(pr);
      };
    }else{
      const char *pbs=*(char **)rec->value;
      long double lstp=1;
      for(int i=0;i<num-1;i++){
	*(pr+i)=atof(pbs);
	lstp-=*(pr+i);
	for(;*pbs&&*pbs!=',';pbs++);
	pbs++;
      };
      for(int i=0;i<num-1;i++){
	*(pr+i)/=lstp;
      };
      ((Mixtreelikelihood *)ff)->setprob(pr);
      ((Mixtreelikelihood *)ff)->setvarprobs(0);
    };
    delete[] pr;
  }else{
    ff=new treelikelihood(a,data);
  };
  if(vars.seektag("-mixture")==NULL){//Simplemodel
    ff->orthpars();
  };
  rec=vars.seektag("-usematrix");
  if(rec!=NULL){
    Realmatrix Q(Qmat);
    params::usematrix(Q);
    ff->setpars(Q);      
  }else{
    rec=vars.seektag("-initpars");
    if(rec!=NULL){
      char *x=new char[50+strlen(*(char **)rec->value)];
      sprintf(x,"Reading parameters from file \"%s\".",*(char**)rec->value);
      info(x,msgcode(1,0));
      ifstream in;
      setifstream(*(char**)rec->value,&in);
      if(!in.good()){
	sprintf(x,"Could not open initial parameter file \"%s\"",*(char**)rec->value);
	recoverableError(x);//should give file menu.
      };
      ff->setpars(in);
      in.close();
      delete[] x;
    }else{
      Realmatrix Q(Qmat);
      ff->setpars(Q);
    };
    if(vars.seektag("-empiricalpi")!=NULL){
      ff->empirical();
    };
  };
  rec=vars.seektag("-empirical");
  if(rec!=NULL){
    const char *empname=*(char **)rec->value;
    int type=0;
    if(!strcmp(empname,"F3x4")){
      type=1;
    }else if(!strcmp(empname,"F1x4")){
      type=2;
    }else if(!strcmp(empname,"Fequal")){
      type=3;
    };
    ff->empirical(type);
  };
  if(vars.seektag("-justbl")){
    info("Using constant parameters.");
    ff->constpars();
  };
  rec=vars.seektag("-mixfile");
  if(rec!=NULL&&num>1){
    const char* mf=*(const char **)rec->value;
    rec=vars.seektag("-path");
    const char **pt;
    if(rec==NULL){
      pt=NULL;
    }else{
      pt=*(const char ***)rec->value;
    };
    ((Mixtreelikelihood *)ff)->setmixfile(mf,pt);
  };
  rec=vars.seektag("-mixstring");
  if(rec!=NULL&&num>1){
    const char* mf=*(const char **)rec->value;
    ((Mixtreelikelihood *)ff)->setmix(mf);
  };
  rec=vars.seektag("-setfixedpars");
  if(rec!=NULL){
    const char *prs=*(char **)rec->value;
    int prnos=1;
    for(const char *prs2=prs;*prs2;prs2++){
      if(*prs2==','){
	prnos++;
      };
    };
    long double *prr=new long double[prnos];
    for(int i=0;i<prnos;i++){
      *(prr+i)=atof(prs);
      for(;*prs&&*prs!=',';prs++);
      prs++;
    };
    ff->setfixedpars(prr);
    delete[] prr;
  };
  ff->normalisejustpars();
  if(skip==0&&!vars.seektag("-noparsimony")){
    info("Using parsimony to obtain first estimate.");
    ff->parsimony();
  };
  return ff;
};

void runsimulation(const char* basefilename,long unsigned int numfiles, long unsigned int seqlength,const tree &t,ifstream &Qmat){
  params p;
  variable *rec=vars.seektag("-usematrix");
  if(rec!=NULL){
    Realmatrix Q(Qmat);
    params::usematrix(Q);
  }else{
    rec=vars.seektag("-initpars");
    if(rec!=NULL){
      char *x=new char[50+strlen(*(char **)rec->value)];
      sprintf(x,"Reading parameters from file \"%s\".",*(char**)rec->value);
      info(x,msgcode(1,0));
      ifstream in;
      setifstream(*(char**)rec->value,&in);
      if(!in.good()){
	sprintf(x,"Could not open initial parameter file \"%s\"",*(char**)rec->value);
	recoverableError(x);//should give file menu.
      };
      p.readpars(in);
      in.close();
      delete[] x;
    }else{
      Realmatrix Q(Qmat);
      p.getpars(Q);
    };
  };
  /*  SHOULD IMPLEMENT THIS FOR MIXTURE MODELS
  rec=vars.seektag("-mixfile");
  if(rec!=NULL&&num>1){
    const char* mf=*(const char **)rec->value;
    rec=vars.seektag("-path");
    const char **pt;
    if(rec==NULL){
      pt=NULL;
    }else{
      pt=*(const char ***)rec->value;
    };
    ((Mixtreelikelihood *)ff)->setmixfile(mf,pt);
  };
  rec=vars.seektag("-mixstring");
  if(rec!=NULL&&num>1){
    const char* mf=*(const char **)rec->value;
    ((Mixtreelikelihood *)ff)->setmix(mf);
    };*/
  if(vars.seektag("-unnormalised")==NULL){
    p.normalise();
  };
  simulate(basefilename,numfiles,seqlength,t,p);
};


void outputnoconverge(int numtimes,treelk *ff,long double *x,int *valpars=NULL){
  ostream *out=&cout;
  variable *rec=vars.seektag("-output");
  if(rec!=NULL){
    out=new ofstream(*(char **)rec->value);
    if(!((ofstream *)out)->is_open()){
      cout<<"Could not open file "<<*(char **)rec->value<<"\n";
      out=&cout;
    };
  };
  if(valpars!=NULL){
    *(out)<<"Parameters selected:\n";
    for(int i=0;*(valpars+i)>0;i++){
      *(out)<<*(valpars+i)<<"  ";
    };
    *(out)<<"\n\n";
  };
  *(out)<<"Failed to converge after "<<numtimes<<" steps. Currently at:\n\n";
  out->precision(10);
  long double val=ff->quickEvaluate(x);
  ff->normalise();
  ff->printall(*out);
  *(out)<<"\n\n Log likelihood="<<val<<"\n\n";
  delete[] x;
  delete ff;
  if(rec!=NULL){
    ((ofstream *)out)->close();
    delete out;
  };
  exit(0);
};


void printconverge(treelk *ff,long double *x,int *valpars=NULL){
  if(valpars!=NULL){
    cout<<"Parameters selected:\n";
    for(int i=0;*(valpars+i)>0;i++){
      cout<<*(valpars+i)<<"  ";
    };
    cout<<"\n\n";
  };
  cout<<"Converged to:\n\n";
  cout.precision(10);
  ff->normalise();
  ff->printall(cout,valpars);
  ff->getx(x);
  long double val=ff->quickEvaluate(x);
  cout<<"\n\n Log Likelihood="<<val<<"\n\n";
  if(vars.seektag("-printsitelikes")){
    long double *sl=ff->sitelikes();
    cout<<"Site\t\tLikelihood\n\n";
    int sites=ff->numsites();
    for(int i=0;i<sites;i++){
      cout<<i+1<<"\t\t"<<*(sl+i)<<"\n";
    };
    delete[] sl;
  };
  if(vars.seektag("-tstats")){
    ff->tstat(x,cout);
  };
};

void outputconverge(treelk *ff,long double *x,int *valpars=NULL){
  ostream *out=&cout;
  variable *rec=vars.seektag("-output");
  if(rec!=NULL){
    out=new ofstream(*(char **)rec->value);
    if(!((ofstream *)out)->is_open()){
      cout<<"Could not open file "<<*(char **)rec->value<<"\n";
      out=&cout;
    };
  };
  *(out)<<"Converged to:\n\n";
  out->precision(10);
  long double val=ff->quickEvaluate(x);
  ff->normalise();
  ff->getx(x);
  Realmatrix oi=ff->oi(x);//Only really necessary if outputing
			  //something that depends on it.
  val=ff->evaluate(x);//already calculated this when calculating oi.
  ff->printtree(*out);
  (*out)<<"\n\n";
  if(vars.seektag("-blstdev")){
    (*out)<<"Standard deviation of branch lengths:\n\n";
    ff->printtreedev(oi,*out);    
    (*out)<<"\n\n";
  };  
  if(valpars!=NULL){
    (*out)<<"Parameters selected:\n";
    for(int i=0;*(valpars+i)>0;i++){
      (*out)<<*(valpars+i)<<"  ";
    };
    (*out)<<"\n\n";
  };
  long double *tc=ff->truepars();
  if(params::numpars>0){
    (*out)<<"Parameter";
    out->width(25);
    (*out)<<"Coefficient";
    out->width(40);
    (*out)<<"Exponential of Coefficient";
    if(vars.seektag("-tstats")){
      out->width(25);
      (*out)<<"t-statistic";
    };
    (*out)<<"\n\n";
    for(int j=0;j<params::numpars;j++){
      out->width(5);
      if(valpars!=NULL&&j>0){
	(*out)<<*(valpars+j-1);
      }else{
	(*out)<<j;
      };
      out->width(29);
      (*out)<<*(tc+j);
      out->width(30);
      (*out)<<exp(*(tc+j));
      if(vars.seektag("-tstats")){
	out->width(25);
	(*out)<<*(tc+j)/sqrt(*(oi.entries+(oi.sz-params::numpars+j)*(oi.sz+1)));
      };
      (*out)<<"\n";
    };
    (*out)<<"\n\n";
  };
  (*out)<<"\n\n Log Likelihood="<<val<<"\n\n";
  if(vars.seektag("-observedinformation")){
    (*out)<<"Observed information matrix:\n\n"<<oi<<"\n\n";
  };
  if(vars.seektag("-printsitelikes")){
    long double *sl=ff->sitelikes();
    *(out)<<"Site\t\tLikelihood\n\n";
    int sites=ff->numsites();
    for(int i=0;i<sites;i++){
      *(out)<<i+1<<"\t\t"<<*(sl+i)<<"\n";
    };
    delete[] sl;
  };
  delete[] tc;
  delete[] x;
  delete ff;
  if(rec!=NULL){
    ((ofstream *)out)->close();
    delete out;
  };
  if(vars.seektag("-keepstate")==NULL){
    rec=vars.seektag("-state");
    if(rec!=NULL){
      remove(*(char **)rec->value);
    }else{
      remove("statefile");
    };
  };
  exit(0);
};


int lratsignif(long double ll,int n,long double lll,int m){
  static long double chisq[]={0,3.84,5.99,7.82,9.49,11.07,12.59,14.07,15.51,16.92,18.31,19.68,21.03,22.36,23.69,25.00,26.30,27.59,28.87,30.14,31.41,32.67,33.92,35.17,36.42,37.65,38.89,40.11,41.34,42.56,43.77,44.99,46.19,47.40,48.60,49.80,51.00,52.19,53.38,54.57,55.76,56.94,58.12,59.30,60.48,61.66,62.83,64.00,65.17,66.34,67.51,68.67,69.83,70.99,72.15,73.31,74.47,75.62,76.78,77.93,79.08,80.23,81.38,82.53,83.68,84.82,85.97,87.11,88.25,89.39,90.53,91.67,92.81,93.95,95.08,96.22,97.35,98.49,99.62,100.75,101.88,103.01,104.14,105.27,106.40,107.52,108.65,109.77,110.90,112.02,113.15,114.27,115.39,116.51,117.63,118.75,119.87,120.99,122.11,123.23,124.34};
  if(n-m<100){
    //    cout<<n-m<<" degrees of freedom.\n";
    //    cout<<"chi-squared value="<<ll-lll<<" compared to "<<chisq[n-m]<<"\n";
    return (ll-lll>chisq[n-m]);//Should improve this.
  }else{
    return (ll-lll>((n-m)*1.01+23.34));//Should improve this.
  };
};

int debuglevel;

void setdebugoutput(ostream& comm){
  variable *rec=vars.seektag("i");
  if(rec!=NULL){
    int inter=atoi(*(char**)rec->value);
    setinteractive(inter);
    if(inter==-1){
      rec=vars.seektag("-logfile");
      if(rec!=NULL){
	setlogfile(*(char **)rec->value);
      }else{
	setlogfile(MAKESTRING(DEFAULTLOG));//Should probably do something cleverer.
      };
    };
  };
  rec=vars.seektag("D");
  debuglevel=0;
  if(rec!=NULL){
    debuglevel=atoi(*(char**)rec->value);
    setdebug(debuglevel);
    if(debuglevel>0){
      comm<<vars;
    };
  };
};

int main(int argc,char *argv[]){
  int op=MAXIMISE;
  int numtimes=3;
  long double eps=1e-6;
  int b=0;

  ifstream tr;
  ifstream data;
  ifstream Qmat;

  //get arguments using a more flexible input scheme.  
  //  getargs(argc,argv,op,numtimes,parfile,treefile,datafile,matrixfile,eps);
  vars.setupvars();
  vars.addvar("treefile","t",&tr,setifstream,MAKESTRING(DEFAULTTREE),noprnt);
  vars.addvar("parfile","p",&(params::in),setifstream,MAKESTRING(DEFAULTPARAMS),noprnt);
  vars.addvar("datafile","d",&data,setifstream,MAKESTRING(DEFAULTDATA),noprnt);
  vars.addvar("matrixfile","m",&Qmat,setifstream,MAKESTRING(DEFAULTMATRIX),noprnt);
  vars.addvar("numtimes","n",&numtimes,setint,MAKESTRING(DEFAULTNUMTIMES),prntint);
  vars.addvar("operation","o",&op,setint,"1",prntint);//MAXIMISE
  vars.addvar("epsilon","e",&eps,setld,"1e-6",prntld);
  vars.setvals(argc,argv);
  SetUpSignalHandlers(&vars);

  atexit(removevarlist);

  variable *rec=vars.seektag("f");
  if(rec!=NULL){
    char *basefname=*(char **)rec->value;
    chooseifstreams(basefname,2,treeextensions,seqextensions,&tr,&data);
  };
  setmodels();
  setdebugoutput(comm);
  int skip=0;
  checkformask();
  rec=vars.seektag("-recover");
  if(rec!=NULL){
    try{
      recover(*((char**)rec->value));
    }catch(int e){
      if(e==0){
	cerr<<"Unable to open recovery file \""<<(char*)rec->value<<"\"\n";
	exit(1);//Could change this to prompt whether to continue.
      }else{
	cerr<<"Unspecified error opening recovery file \""<<(char*)rec->value<<"\"\n";
	exit(1);//Could change this to prompt whether to continue.
      };
    };
    if(Recover!=NULL){
      skip=getskip(Recover->retid);
    };
  };
  rec=vars.seektag("T");
  if(rec!=NULL){
    setnumthreads(atoi(*(char**)rec->value));
  };
  state s("start");
  rec=vars.seektag("-state");
  if(rec!=NULL){
    s.setfile(*(char**)rec->value);
  }else{
    s.setfile("statefile");
  };
  setstate(&s);
  rec=vars.seektag("b");
  if(rec!=NULL){
    b=atoi(*(char**)rec->value);
  };
  int pa=0;
  rec=vars.seektag("P");
  if(rec!=NULL){
    pa=atoi(*(char**)rec->value);
  };
  rec=vars.seektag("-testparunselect");
  if(rec!=NULL){
    int *sel=selection(*(char **)rec->value);
    testundopars(sel,Qmat);
    exit(0);
  };
  try{
    rec=vars.seektag("-treenumber");
    if(rec!=NULL){
      int tn=atoi(*(char**)rec->value);
      //      cout<<"Selecting tree number "<<tn<<"\n";
      for(int i=0;i<tn;i++){
	delete[] readstring(tr,"\n");//Should improve checking
					     //for blank lines.
	skipblanklines(tr);
      };
    };
    char *a=readstring(tr,"\n");
    tr.close();

    rec=vars.seektag("-simulate");
    if(rec!=NULL){
      data.close();
      char *nsim=*(char **)rec->value;
      int i=0;
      long unsigned int numfiles=1;
      long unsigned int seqlength=0;
      for(;*(nsim+i)&&*(nsim+i)!=','&&*(nsim+i)!=':';i++);
      if(*(nsim+i)==','){
	numfiles=atol(nsim);
	seqlength=atol(nsim+i+1);
      }else{
	seqlength=atol(nsim);
      };
      for(;*(nsim+i)&&*(nsim+i)!=':';i++);
      if(*(nsim+i)==':'){
	srand(atol(nsim+i+1));
      };
      const char *basefilename;
      rec=vars.seektag("-output");
      if(rec!=NULL){
	basefilename=*(char **)rec->value;
      }else{
	basefilename="simulation-output";
      };
      tree t(a);
      delete[] a;
      runsimulation(basefilename,numfiles,seqlength,t,Qmat);
      t.remove();
      Qmat.close();
      exit(0);
    };
    treelk *ff=maketreelikelihood(a,data,Qmat,skip);
    info("Set up starting values.");
    long double *x=new long double[ff->dim];
    if(skip==0){
      ff->getx(x);
    }else{
      Recover->load(1,ff->dim,x);
      ff->quickEvaluate(x);//To set the x values for better debugging.
      //      ff->normalisejustpars();//Otherwise, scaling coefficient will be wrong.
    };
    Qmat.close();
    delete[] a;
    data.close();
    long double initpen=0;
    rec=vars.seektag("-initpen");
    if(rec!=NULL){
      initpen=atof(*(char**)rec->value);
    };
    if(vars.seektag("-testderivs")){
      info("Testing derivatives.");
      testfunction(*ff,x,eps,5e-3);
      exit(0);
    };
    int stv=CurrentState->addvar('L',ff->dim,x,"newx");
    info("Starting from position:\n");
    if(debuglevel>=0){
      ff->printall();
    };
    info("Beginning optimisation.");
    info("Trying Newton-Raphson Method:");

    if(vars.seektag("-variableselect")){
      int *valpars=new int[params::numpars];
      int *oldvalpars=new int[params::numpars];
      for(int i=0;i<params::numpars-1;i++){
	*(valpars+i)=i+1;
      };
      *(valpars+params::numpars-1)=-1;
      long double thresh[]={TSTAT10,TSTAT5,TSTAT25,TSTAT1,TSTAT05,TSTAT01};
      int thr=0;
      int maxthr=6;
      long double *oldx=new long double[ff->dim];
      int *tselold=NULL;
      long double lastlike=1;
      int lastnum=params::numpars;
      long double tolerance=0.02;
      for(;thr<maxthr;){
	if(!NewtonRaphson(*ff,x,numtimes,op,stv,ff->posvals(),tolerance,initpen)){
	  outputnoconverge(numtimes,ff,x,valpars);
	};
	tolerance=1e-4;
	long double val=ff->evaluate(x);
	if(lastlike<=0&&lratsignif(lastlike,lastnum,val,params::numpars)){
	  ff->undoselect(tselold,lastnum);
	  outputconverge(ff,oldx,oldvalpars);
	};
	int nochange;
	printconverge(ff,oldx,valpars);
	ff->normalise();
	ff->getx(x);
	do{
	  nochange=0;
	  //	  cout<<"Threshold="<<thresh[thr]<<"\n";
	  int *tsel=ff->tstatsel(x,thresh[thr]);
	  int i=0;
	  for(int i=0;i<params::numpars-1;i++){
	    *(oldvalpars+i)=*(valpars+i);
	  };
	  *(oldvalpars+params::numpars-1)=-1;
	  for(;*(tsel+i)>0;i++){
	    *(valpars+i)=*(valpars+*(tsel+i)-1);
	  };
	  *(valpars+i)=-1;//*(valpars+*(tsel+i)-1);
	  //	  for(int j=0;*(valpars+j)>0;j++){
	  //	    cout<<*(valpars+j)<<"  ";
	  //	  };
	  //	  cout<<"\n\n";
	  if(i==params::numpars-1){//No change
	    thr++;
	    nochange=1;
	  }else{
	    lastnum=params::numpars;
	    for(int j=0;j<ff->dim;j++){
	      *(oldx+j)=*(x+j);
	    };
	    ff->normalise();
	    ff->selectpars(tsel);
	    ff->normalisejustpars();
	    ff->getx(x);
	    if(debuglevel>=0){
	      for(int i=0;*(tsel+i)>=0;i++){
		cout<<*(tsel+i)<<"  ";
	      };
	      cout<<"\n";
	      ff->printall();
	    };
	  };
	  if(tselold!=NULL){
	    delete[] tselold;
	  };
	  tselold=tsel;
	}while(nochange&&thr<maxthr);
	lastlike=val;
      };
      outputconverge(ff,x,valpars);
    }else{
      long double tolerance=1e-5;
      if(NewtonRaphson(*ff,x,numtimes,op,stv,ff->posvals(),tolerance,initpen)){
	outputconverge(ff,x);
      }else{
	outputnoconverge(numtimes,ff,x);
      };
    };
  }catch(int u){
    cerr<<"Badly formatted tree file.\n"<<u<<" left brackets not closed.";
    exit(1);
  };
};
