/*

   Copyright 2011 Toby Kenney

   This file is part of Cold.

   Cold is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   Cold is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with Cold.  If not, see <http://www.gnu.org/licenses/>.

 */

#ifndef MIXTREELIKELIHOOD_H
#define MIXTREELIKELIHOOD_H

#include "Tree.h"
#include "Parameters.h"
#include "Sequence.h"
#include "Matrix.h"
#include "Optimisation.h"
#include "TreeLikelihood.h"

// Code needed for optimisation.

class Mixture{
  //determines which parameters are the same across all models 
  //and which parameters are constant.
  int num;
  int npars;
  int *pars;// Negative values indicate a continuation of the previous
	    // parameter. List is terminated by 0.
public:
  Mixture(){num=0;npars=0;pars=NULL;};
  Mixture(int n){num=n;npars=0;pars=NULL;};
  Mixture(const char *s,int n);
  Mixture(istream& s,int n);
  ~Mixture(){if(pars!=NULL){delete[] pars;};};
  int valid(){return pars!=NULL;};
  void setnum(int n){num=n;};
  void assign(const char *s);
  void assign(istream &s);
  friend class Mixtreelikelihood;
  void setpars(const long double *vals,long double *in);
  void sumpars(const long double *in,long double *out);
  void sumparstrisquare(const long double *dt,long double *to,int dim,int offset=0);
};

class Mixtreelikelihood : public treelk{
  tree T;
  int num;
  params *p;
  long double *prob;
  int varprobs;
  sequence *d;
  Factmatrix *f;
  long double *lastval;
  long double *vals;
  int *variablepars;//Makes a list of parameters allowed to vary.
  int dir;
  Mixture m;
  int testlast(const long double *x);
  Mixtreelikelihood(const Mixtreelikelihood& T);//Disable copying
  inline void calculatevals(const long double *x);
public:
  Mixtreelikelihood(int mx,const char *tr,const Realmatrix& Q,ifstream& data);
  Mixtreelikelihood(int mx,const char *tr,ifstream& data);
  void setpars(const Realmatrix& Q);
  void setpars(ifstream& I);
  ~Mixtreelikelihood(){delete[] d;delete[] p;delete[] prob;T.up=NULL;T.remove();for(int i=0;i<num;i++){(f+i)->remove();};delete[] f;delete[] lastval;if(vals!=NULL){delete[] vals;};delete[] variablepars;};
  void readpars(istream &in){for(int i=0;i<num;i++){(p+i)->readpars(in);*(f+i)=(p+i)->fmatrix();};};
  void getx(long double *out){for(int i=0;i<dim+1-T.edges()-num;i++){*(out+i+T.numed)=(p+*(variablepars+i)/params::numpars)->coeff[*(variablepars+i)%params::numpars];};for(int i=0;i<num-1;i++){*(out+dim+1-num+i)=*(prob+i);};T.putedgelengths(out);};
  long double evaluate(const long double *x);
  long double *sitelikes(){return NULL/*T.Mixsitelikelihoods(f,d,p)*/;};
  int numsites(){return d->length;};
  void deriv(const long double *x,long double *out);
  void hessian(const long double *x,long double *out);
  /*  long double directionalDeriv(const long double *x,const long double *v);
  long double directionalSecDer(const long double *x,const long double *v);
  void approxHessian(const long double *x,long double *out);
  int LineSearch(long double *x,long double lastval,const long double *dir,int numpos);*/
  long double quickEvaluate(const long double *x){return this->evaluate(x);};
  //Should make this better.
  int edges(){return T.edges();};
  int posvals(){return T.numed+varprobs;};
  void constpars(){dim=T.edges();*variablepars=-1;};
  void printtree(ostream& out=cout){T.print(out);};
  void printdata(){T.printdata(cout,d);};
  void normalise();
  void normalisejustpars();
  void orthpars(){params::orthmats();};
  void printall(ostream& out=cout,int *valpars=NULL);
  void parsimony();
  void empirical(int type=0){long double *epi=emppi(d,T.leaves(),type);for(int i=0;i<num;i++){(p+i)->setpi(epi);};/*cout<<"Empirical frequencies:\n";for(int j=0;j<NumCodons;j++){cout<<*(epi+j)<<"\t";};cout<<"\n";*/ delete[] epi;};
  void setprob(long double *pr){for(int i=0;i<num-1;i++){*(prob+i)=*(pr+i);};};
  void setmix(const char*s){m.assign(s);dim=T.numed+num-1+m.npars;};
  void setmix(ifstream& s){m.assign(s);dim=T.numed+num-1+m.npars;};
  void setmixfile(const char* s,const char *const*pt=NULL);
  void setvarprobs(int i){dim+=varprobs-i;varprobs=i;};
  void setfixedpars(long double *vls);
  void tstat(long double *x,ostream& out);
  int* tstatsel(long double *x,long double threshold=TSTAT5);
  void selectpars(int *sel);
  void undoselect(int *sel,int lastpars);
  void printtreedev(const Realmatrix &oi,ostream& out=cout){long double *std=new long double[T.numed]; for(int i=0;i<T.numed;i++){*(std+i)=sqrt(*(oi.entries+i*(oi.sz+1)));};T.print(std,out);delete[] std;};
  long double *truepars(int *valpars=NULL);
  RealmatrixT oi(long double *x);
};

#endif
